#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from enum import Enum

from pydantic import AnyUrl, BaseModel, ConfigDict, Field, StrictInt, StrictStr


class JobStatusCompleted(Enum):
    failed = "failed"
    canceled = "canceled"
    succeeded = "succeeded"


class JobStatusOngoing(Enum):
    requested = "requested"
    in_progress = "in progress"
    canceling = "canceling"


class JobStatusEnum(Enum):
    """
    Enum representing the status of a job.
    """

    requested = "requested"
    in_progress = "in progress"
    succeeded = "succeeded"
    failed = "failed"
    cancelling = "cancelling"
    cancelled = "cancelled"


class CompletedJobLinks(BaseModel):
    results: AnyUrl


class OngoingJobLinks(BaseModel):
    cancel: AnyUrl


class ExecuteTaskRequest(BaseModel):
    parameters: dict[str, StrictStr]


class OngoingJobResponse(BaseModel):
    status: JobStatusOngoing


class Error(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )

    status: StrictInt
    """The status code of the error."""

    type_: StrictStr = Field(alias="type")
    """The type of the error."""

    title: StrictStr
    """The title of the error."""

    detail: StrictStr | None = None
    """A message describing the error."""


class JobStatusResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )

    status: JobStatusEnum
    """The status of the job."""

    progress: StrictInt | None = Field(None, ge=0, le=100)
    """A number between 0 and 100 representing the progress of the job."""

    message: StrictStr | None = None
    """A message describing the current progress of the job."""

    error: Error | None = None
    """An error that occurred during the job."""


class CompletedJobResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )

    status: JobStatusEnum
    """The status of the job."""

    results: dict | None = None
    """The results of the job."""

    error: Error | None = None
    """An error that occurred during the job."""



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Task API
=============


    The Task API provides the ability to execute computation tasks in Evo that require variable, on-demand
    processing power. Building the Task API into your application can enable fast processing of long
    running, or resource intensive operations, without depending on the end user's physical hardware.

    Tasks are created, triggering a job to be executed asynchronously within a specific topic for your organization,
    and can be monitored throughout their execution lifecycle.

    For more information on using the Task API, see [Overview](/docs/guides/tasks), or the API
    references here.

This code is generated from the OpenAPI specification for Task API.
API version: 0.0.2
"""

# Import endpoint apis.
from .api import JobsApi, TasksApi

__all__ = [
    "JobsApi",
    "TasksApi",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import copy

from evo.common.exceptions import BaseTypedError, EvoClientException


class JobError(BaseTypedError):
    """Raised when a job fails."""

    def copy(self) -> JobError:
        """Return a copy of the error.

        Job result responses are cached so that they can be retrieved multiple times without
        re-fetching. This method is used to create a copy of an error result so that it can be
        raised from a cached copy without modifying the original.

        :return: A copy of the error.
        """
        return JobError(
            status=self.status,
            reason=self.reason,
            content=copy.deepcopy(self.content),
            headers=self.headers.copy(),
        )


class JobPendingError(EvoClientException):
    """Raised when results are requested but the job is still pending."""

    def __init__(self, url: str, status: str) -> None:
        super().__init__(f"Job at {url} is still pending with status: {status}")



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Task API
=============


    The Task API provides the ability to execute computation tasks in Evo that require variable, on-demand
    processing power. Building the Task API into your application can enable fast processing of long
    running, or resource intensive operations, without depending on the end user's physical hardware.

    Tasks are created, triggering a job to be executed asynchronously within a specific topic for your organization,
    and can be monitored throughout their execution lifecycle.

    For more information on using the Task API, see [Overview](/docs/guides/tasks), or the API
    references here.

This code is generated from the OpenAPI specification for Task API.
API version: 0.0.2
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod
from evo.common.utils import get_header_metadata

from ..models import *  # noqa: F403

__all__ = ["TasksApi"]


class TasksApi:
    """API client for the Tasks endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def execute_task(
        self,
        org_id: str,
        topic: str,
        task: str,
        execute_task_request: ExecuteTaskRequest | None = None,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Trigger an asynchronous execution of a task

        Trigger an asynchronous task within a specific topic for a designated organization. The task, topic, and organization ID are supplied as path parameters. A JSON request body, referencing the **ExecuteTaskRequest** schema and containing task-specific parameters, is required for this operation. Upon successful start, a 303 response is returned. The URL of the status endpoint for the newly created job is indicated in the Location header of the response.

        :param org_id: The organization identifier
            Format: `uuid`
            Example: `'org_id_example'`
        :param topic: The compute topic
            Example: `'topic_example'`
        :param task: The compute task
            Example: `'task_example'`
        :param execute_task_request: (optional)
            Example: `endpoints.ExecuteTaskRequest()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """  # noqa: E501
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "topic": topic,
            "task": task,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "303": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/compute/orgs/{org_id}/{topic}/{task}",
            path_params=_path_params,
            header_params=_header_params,
            body=execute_task_request,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
import copy
import json
import re
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from typing import Any, Generic, TypeVar
from uuid import UUID

from evo.common import APIConnector, HTTPResponse
from evo.common.exceptions import UnknownResponseError
from evo.common.interfaces import IFeedback
from evo.common.utils import NoFeedback, Retry
from pydantic import TypeAdapter, ValidationError

from evo import logging

from .data import JobProgress, JobStatusEnum
from .endpoints import JobsApi, TasksApi
from .endpoints.models import CompletedJobResponse
from .exceptions import JobError, JobPendingError

logger = logging.getLogger("compute.client")

__all__ = [
    "JobClient",
]

T_Result = TypeVar("T_Result")


@contextmanager
def _validating_pydantic_model(ctx: HTTPResponse) -> Iterator[None]:
    """Context manager to handle Pydantic model validation errors.

    :param ctx: The HTTP response context.

    :raises UnknownResponseError: If the managed context raises a pydantic ValidationError.
    """
    try:
        yield
    except ValidationError as e:
        logger.error(f"Failed to validate response: {e}", exc_info=e)

        content = ctx.data
        try:
            content = content.decode("utf-8")  # Raises UnicodeDecodeError.
            content = json.loads(content)  # Raises json.JSONDecodeError.
        except (UnicodeDecodeError, json.JSONDecodeError):
            pass  # content will fall back to the most specific valid value.

        raise UnknownResponseError(
            status=ctx.status,
            reason=ctx.reason,
            content=content,
            headers=ctx.headers,
        ) from e


RE_STATUS_URL = re.compile(r"compute/orgs/(?P<org_id>[^/]+)/(?P<topic>[^/]+)/(?P<task>[^/]+)/(?P<job_id>[^/]+)/status")


class JobClient(Generic[T_Result]):
    """Client for managing a submitted compute task (a job)."""

    def __init__(
        self,
        connector: APIConnector,
        org_id: UUID,
        topic: str,
        task: str,
        job_id: UUID,
        result_type: type[T_Result] = dict,
        preview: bool = False,
    ) -> None:
        """
        :param connector: The API connector to use.
        :param org_id: The organization ID.
        :param topic: The topic of the task.
        :param task: The task to be executed.
        :param job_id: The job ID.
        :param result_type: The type to validate the result against.
        :param preview: Whether to use preview mode and include necessary header information.
        """
        self._connector = connector
        self._org_id = org_id
        self._topic = topic
        self._task = task
        self._job_id = job_id
        self._url = connector.base_url + f"compute/orgs/{self._org_id}/{self._topic}/{self._task}/{self._job_id}/status"
        self._type_adapter = TypeAdapter(result_type)
        self._mutex = asyncio.Lock()
        self._results: T_Result | JobError | None = None
        self._preview = preview

    @property
    def id(self) -> UUID:
        """The job id."""
        return self._job_id

    @property
    def topic(self) -> str:
        """The topic of the task."""
        return self._topic

    @property
    def task(self) -> str:
        """The task."""
        return self._task

    @property
    def url(self) -> str:
        """The status URL of the job."""
        return self._url

    def __repr__(self) -> str:
        return self.url

    def _get_headers(self) -> dict[str, str]:
        """Get the headers dictionary for API requests.

        :return: Headers dictionary with preview header if enabled.
        """
        return {"API-Preview": "opt-in"} if self._preview else {}

    @staticmethod
    def from_url(
        connector: APIConnector, url: str, result_type: type[T_Result] = dict, preview: bool = False
    ) -> JobClient[T_Result]:
        """Create a job client from a status URL.

        The URL hostname must match the connector base URL.

        :param connector: The API connector to use.
        :param url: The status URL of a submitted job.
        :param result_type: The type to validate the result against.
        :param preview: Whether to use preview mode and include necessary header information.

        :return: A client for managing the referenced job.
        """
        if not url.startswith(connector.base_url):
            raise ValueError("The job URL does not match the connector base URL.")

        if match := RE_STATUS_URL.fullmatch(url.removeprefix(connector.base_url)):
            path_params = match.groupdict()
            for key in path_params:
                if key.endswith("_id"):
                    try:
                        path_params[key] = UUID(path_params[key])
                    except ValueError:
                        raise ValueError(f"Invalid {key.removesuffix('_id')} ID in URL: {url}") from None

        return JobClient(connector=connector, **path_params, result_type=result_type, preview=preview)

    @staticmethod
    async def submit(
        connector: APIConnector,
        org_id: UUID,
        topic: str,
        task: str,
        parameters: Mapping[str, Any],
        result_type: type[T_Result] = dict,
        preview: bool = False,
    ) -> JobClient[T_Result]:
        """Trigger an asynchronous task within a specific topic with the given parameters.

        :param connector: The API connector to use.
        :param topic: The topic of the task.
        :param task: The task to be executed.
        :param parameters: The parameters for the task.
        :param result_type: The type to validate the result against.
        :param preview: Whether to use preview mode and include necessary header information.

        :return: The job that was created.

        :raises UnknownResponseError: If the Location header is missing or invalid.
        """

        async with connector:
            response = await TasksApi(connector).execute_task(
                org_id=str(org_id),
                topic=topic,
                task=task,
                execute_task_request={"parameters": dict(parameters)},
                additional_headers={"API-Preview": "opt-in"} if preview else {},
            )

        # Location header is the status endpoint of the created job.
        # i.e., compute/orgs/{org_id}/{topic}/{task}/{job_id}/status
        try:
            job_url = response.headers["Location"]
            job_url = connector.base_url + job_url.removeprefix(connector.base_url).removeprefix("/")
            return JobClient.from_url(connector, job_url, result_type, preview=preview)
        except (KeyError, ValueError):
            raise UnknownResponseError(
                status=response.status, reason=response.reason, content=None, headers=response.headers
            )

    async def get_status(self) -> JobProgress:
        """Get the status of the job.

        :return: The job progress.
        """
        async with self._connector:
            response = await JobsApi(self._connector).get_job_status(
                org_id=self._org_id,
                topic=self._topic,
                task=self._task,
                job_id=self._job_id,
                additional_headers=self._get_headers(),
            )

        if response.error:
            error = JobError(
                status=response.error.status,
                reason=None,
                content=response.error.model_dump(by_alias=True, exclude_unset=True, exclude_defaults=True),
                headers=None,
            )
        else:
            error = None

        return JobProgress(
            status=JobStatusEnum(response.status.value),
            progress=response.progress,
            message=response.message,
            error=error,
        )

    async def _get_results(self) -> T_Result | JobError:
        """Get the results of the job if they are not already cached.

        The results are cached after the first successful call.

        :return: The job results or error.

        :raises UnknownResponseError: If the response or results fail validation.
        :raises JobPendingError: If the job is still pending.
        """
        async with self._mutex:
            if self._results is None:
                # Request the results from the API.
                async with self._connector:
                    response = await JobsApi(self._connector).get_job_results(
                        org_id=self._org_id,
                        topic=self._topic,
                        task=self._task,
                        job_id=self._job_id,
                        additional_headers=self._get_headers(),
                    )

                with _validating_pydantic_model(response):
                    job = CompletedJobResponse.model_validate_json(response.data)

                if response.status == 202:
                    # The job is still pending, raise an error.
                    raise JobPendingError(url=self.url, status=job.status.value)

                if error := job.error:
                    # The job failed, save the error.
                    self._results = JobError(
                        status=error.status,
                        reason=None,
                        content=error.model_dump(by_alias=True, exclude_unset=True, exclude_defaults=True),
                        headers=response.headers,
                    )
                else:
                    # The job succeeded, deserialize and save the results.
                    with _validating_pydantic_model(response):
                        # Nitpick: job.results _may_ in fact be `None`, in which case the type adapter will
                        # raise ValidationError and trigger the error handling in the context manager.
                        # At the time of writing, it is unclear whether there are any valid cases where both
                        # the error and results are missing when the job is in a completed state.
                        self._results = self._type_adapter.validate_python(job.results)

            # Return a copy of the cached results.
            if isinstance(self._results, JobError):
                return self._results.copy()
            else:
                return copy.deepcopy(self._results)

    async def get_results(self) -> T_Result:
        """Get the results of the job.

        :return: The results.

        :raises UnknownResponseError: If the response or results fail validation.
        :raises JobPendingError: If the job is still pending.
        :raises JobError: If the job failed.
        """
        if isinstance(results := await self._get_results(), JobError):
            raise results.with_traceback(None)
        return results

    async def cancel(self) -> None:
        """Cancel the job."""
        async with self._connector:
            await JobsApi(self._connector).cancel_job(
                org_id=self._org_id,
                topic=self._topic,
                task=self._task,
                job_id=self._job_id,
                additional_headers=self._get_headers(),
            )

    async def wait_for_results(
        self, polling_interval_seconds: float = 0.5, retry: Retry | None = None, fb: IFeedback = NoFeedback
    ) -> T_Result:
        """Wait for the job to complete and return the results.

        :param polling_interval_seconds: The interval in seconds between status checks.
        :param retry: A Retry object with a wait strategy. If None, a default Retry is created.
        :param fb: The feedback object to use.

        :return: The results.

        :raises UnknownResponseError: If the response or results fail validation.
        :raises JobPendingError: If the job is still pending.
        :raises JobError: If the job failed.
        """
        if retry is None:
            retry = Retry(logger)

        latest_progress = 0.0
        latest_message = "Waiting on remote job..."

        while True:
            async for handler in retry:
                with handler.suppress_errors():
                    latest = await self.get_status()

            if latest.status in (JobStatusEnum.succeeded, JobStatusEnum.failed, JobStatusEnum.cancelled):
                break

            # Task API returns progress as an integer between 0 and 100
            latest_progress = latest_progress if latest.progress is None else latest.progress * 0.01
            latest_message = latest_message if latest.message is None else latest.message

            fb.progress(latest_progress, latest_message)
            await asyncio.sleep(polling_interval_seconds)

        fb.progress(1.0, "Fetching results...")
        return await self.get_results()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from dataclasses import dataclass

from .endpoints.models import JobStatusEnum
from .exceptions import JobError

__all__ = [
    "JobProgress",
    "JobStatusEnum",
]


@dataclass(frozen=True, kw_only=True)
class JobProgress:
    status: JobStatusEnum
    """The status of the job."""

    progress: int | None = None
    """A number between 0 and 100 representing the progress of the job."""

    message: str | None = None
    """A message describing the current progress of the job."""

    error: JobError | None = None
    """An error associated with the job, if any."""

    def __str__(self) -> str:
        msg = f"[{self.status.value}]"
        if self.progress is not None:
            msg += f" {self.progress}%"
        if self.message is not None:
            msg += f" > {self.message}"
        if self.error is not None:
            msg += f"\n{self.error}"
        return msg
